#ifndef _FTL_FILTERS_FILTER_HPP_
#define _FTL_FILTERS_FILTER_HPP_

#include <list>
#include <ftl/configurable.hpp>
#include <ftl/configuration.hpp>
#include <ftl/rgbd/frame.hpp>
#include <ftl/rgbd/source.hpp>
#include <ftl/cuda_common.hpp>

namespace ftl {
//namespace filters {

/**
 * An abstract frame filter interface. Any kind of filter that operates on a
 * single frame should use this as a base class. An example of a filter would
 * be MLS smoothing, or optical flow temporal smoothing. Some 'filters' might
 * simply generate additional channels, such as a 'Normals' filter that
 * generates a normals channel. Filters may also have internal data buffers,
 * these may also persist between frames in some cases.
 */
class Filter : public ftl::Configurable {
	public:
	explicit Filter(nlohmann::json &config);
    virtual ~Filter();

	virtual void filter(ftl::rgbd::Frame &f, ftl::rgbd::Source *s, cudaStream_t stream)=0;

	inline void enable() { enabled_ = true; }
	inline void disable() { enabled_ = false; }
	inline bool enabled() const { return enabled_; }
	inline void enabled(bool e) { enabled_ = e; }

	protected:
	bool enabled_;
};

/**
 * Represent a sequential collection of filters. Each filter created is
 * added to an internal list and then applied to a frame in the order they were
 * created.
 */
class Filters : public ftl::Configurable {
	public:
	explicit Filters(nlohmann::json &config);
    ~Filters();

	template <typename T>
	Filter *create(const std::string &name);

	void filter(ftl::rgbd::Frame &f, ftl::rgbd::Source *s, cudaStream_t stream=0);

	private:
	std::list<ftl::Filter*> filters_;
	Filter *_create(Filter *f);
};

//}
}

template <typename T>
ftl::Filter *ftl::Filters::create(const std::string &name) {
	return _create(dynamic_cast<ftl::Filter*>(ftl::create<T>(this, name)));
}

#endif  // _FTL_FILTERS_FILTER_HPP_
